# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-06 15:29

@author: a002028
"""
from bokeh.models import (
    TextInput,
    ColumnDataSource,
    Circle,
    TapTool,
    HoverTool,
    WheelZoomTool,
    ResetTool,
    PanTool,
    SaveTool,
    LassoSelectTool,
    ColorBar,
    LinearColorMapper
)
from bokeh.layouts import grid, row, column, Spacer
from bokeh.models.widgets import Select, RangeSlider, DataTable, TableColumn, Panel, Tabs
from bokeh.plotting import figure, show
from bokeh.tile_providers import get_provider, Vendors
from bokeh.core.validation import silence
from bokeh.core.validation.warnings import FIXED_SIZING_MODE
import numpy as np
import pandas as pd
from matplotlib import colors
from matplotlib import cm
from ctdvis.utils import get_time_as_format, convert_projection, get_contour_data
from ctdvis.callbacks import cbs
from ctdvis.sources.ts import Source as TS_Source
from ctdvis.sources.data import setup_data_source
from ctdvis.widgets.paragraph import (
    standard_block_header,
    get_info_block,
    get_export_info_block
)
silence(FIXED_SIZING_MODE, True)


class QCWorkTool:
    """The almighty QC-tool.

    Used for quality control of high resolution CTD-data (oceanography).
    """

    # TODO: Well, this is great.. However, we need to simplify and
    #  divide this class into widgets instead.. to be continued..
    #  - delete color fields? use flag fields instead?

    def __init__(self,
                 dataframe,
                 datasets=None,
                 parameters=None,
                 plot_keys=None,
                 color_fields=None,
                 qflag_fields=None,
                 auto_q_flag_parameters=None,
                 tabs=None,
                 plot_parameters_mapping=None,
                 ctdpy_session=None,
                 multi_sensors=False,
                 combo_plots=False,
                 output_filename="CTD_QC_VIZ.html",
                 output_as_notebook=False,
                 user_download_directory=None,
                 ):
        """Initiate."""
        self.seconds = ColumnDataSource(data=dict(tap_time=[None], reset_time=[None]))
        self.ctd_session = ctdpy_session
        self.multi_sensors = multi_sensors
        self.combo_plots = combo_plots

        self.map = None
        self.plot_keys = plot_keys
        self.datasets = datasets
        self.key_ds_mapper = self.get_mapper_key_to_ds()
        self.parameters = parameters
        self.plot_parameters_mapping = plot_parameters_mapping
        self.color_fields = color_fields
        self.qflag_fields = qflag_fields
        self.auto_qflag_fields = auto_q_flag_parameters
        self.tabs = tabs
        self.output_as_notebook = output_as_notebook
        if self.output_as_notebook:
            raise NotImplementedError('Not yet applicable to work with notebooks!')
            # output_notebook()
        # else:
        #     output_file(output_filename)

        self.tile_provider = get_provider(Vendors.CARTODBPOSITRON_RETINA)

        self.figures = {}
        xrange_callbacks = {}
        y_range_setting = None
        for p in self.plot_parameters_mapping:
            if p == 'y' or 'q' in p or p[0].isupper() or p.startswith('color'):
                continue
            param = self.plot_parameters_mapping.get(p)
            self.figures[p] = figure(tools="pan,reset,wheel_zoom,lasso_select,save",
                                     active_drag="lasso_select",
                                     title="", height=400, width=400,
                                     y_range=y_range_setting,
                                     tooltips=[(param, "@{}".format(p)),
                                               ("Pressure [dbar]", "@y"),
                                               ("Auto-QC", "@{}_q0".format(p))])
            self.figures[p].title.align = 'center'
            self.figures[p].xaxis.axis_label = param
            self.figures[p].xaxis.axis_label_text_font_style = 'bold'
            self.figures[p].ygrid.band_fill_alpha = 0.05
            self.figures[p].ygrid.band_fill_color = "black"
            self.figures[p].toolbar.active_scroll = self.figures[p].select_one(WheelZoomTool)

            if not y_range_setting or (self.multi_sensors and p == 'x4'):
                self.figures[p].yaxis.axis_label = 'Pressure (dbar)'
                self.figures[p].yaxis.axis_label_text_font_style = 'bold'

            y_range_setting = y_range_setting or self.figures[p].y_range

            xrange_callbacks[p] = cbs.x_range_callback(x_range_obj=self.figures[p].x_range,
                                                       seconds=self.seconds)
            self.figures[p].x_range.js_on_change('start', xrange_callbacks[p])

        if self.combo_plots and self.multi_sensors:
            # TODO check the p2 (_) variable.. Should it not be used?
            for name, p1, _ in zip(
                    ('COMBO_TEMP', 'COMBO_SALT', 'COMBO_DOXY'),
                    ('x1', 'x2', 'x3'), ('x4', 'x5', 'x6')):
                param = self.plot_parameters_mapping.get(p1)
                self.figures[name] = figure(tools="pan,reset,wheel_zoom,lasso_select,save",
                                            active_drag="lasso_select",
                                            title="", height=400, width=400,
                                            y_range=y_range_setting,
                                            )
                self.figures[name].title.align = 'center'
                self.figures[name].xaxis.axis_label = param
                self.figures[name].xaxis.axis_label_text_font_style = 'bold'
                self.figures[name].ygrid.band_fill_alpha = 0.05
                self.figures[name].ygrid.band_fill_color = "black"
                self.figures[name].toolbar.active_scroll = \
                    self.figures[name].select_one(WheelZoomTool)

                if p1 == 'x1':
                    self.figures[name].yaxis.axis_label = 'Pressure (dbar)'
                    self.figures[name].yaxis.axis_label_text_font_style = 'bold'

        cbs.add_hlinked_crosshairs(*(fig_obj for i, fig_obj in self.figures.items()))

        self.ts = figure(
            title="CTD TS Diagram", tools=[PanTool(), WheelZoomTool(), ResetTool(), SaveTool()],
            tooltips=[("Serie", "@key")], height=400, width=400,
            x_range=(2, 36), y_range=(-2, 20)
        )
        self.ts.title.align = 'center'

        self._setup_position_source(dataframe)
        self.data_source = setup_data_source(
            dataframe,
            pmap=self.plot_parameters_mapping,
            key_list=np.unique(self.position_source.data['KEY']),
            parameter_list=self.color_fields + self.plot_keys + self.auto_qflag_fields + ['COMNT_SAMP']  # noqa: E501
        )

        self.ts_source = TS_Source()
        self.ts_source.setup_source(dataframe, self.plot_parameters_mapping)
        self.ts_plot_source = ColumnDataSource(data=dict(x=[], y=[], color=[], key=[]))

        self._setup_month_selector()
        self._setup_comnt_inputs()
        self._setup_selection_widgets()
        self._setup_multiflag_widget()
        self._setup_flag_widgets()
        self._setup_reset_callback(**xrange_callbacks)
        self._setup_datasource_callbacks()
        self._setup_download_button(user_download_directory)
        self._setup_get_file_button()
        self._setup_serie_table()
        self._setup_info_block()
        self._setup_map()

        self.ts_axis_ranges = {'t_min': 0, 't_max': 25, 's_min': 2, 's_max': 36}

    def get_mapper_key_to_ds(self):
        """Return mapper between dataset filename and key.

        Key = serie key (eg. '20191009_77SE_0005')
        """
        # TODO: would we like to create this mapper in any other way?
        #  Let´s say that the dataset name doesn't starts with "ctd_profile_"
        #  mapper = {}
        #  for key, item in self.datasets.items():
        return {ds_name.replace('ctd_profile_', '').replace('.txt', ''): ds_name
                for ds_name in self.datasets}

    @staticmethod
    def _get_monthly_keys(position_df):
        """Return mapper of monthly keys."""
        # FIXME Is this really necessary ?
        dictionary = {'All': position_df['KEY'].to_list()}
        for month in [str(m).zfill(2) for m in range(1, 13)]:
            boolean = position_df['MONTH'] == month
            dictionary[month] = position_df.loc[boolean, 'KEY'].to_list()
        return dictionary

    def _setup_position_source(self, df):
        """Set position data sources.

        self.position_source: Contains all position information.
        self.position_plot_source: Contains only the selected information.
        """
        position_df = df[
            ['STATION', 'LATITUDE_DD', 'LONGITUDE_DD', 'KEY', 'MONTH']
        ].drop_duplicates(keep='first').reset_index(drop=True)

        for pos_row in position_df.itertuples():
            try:
                float(pos_row.LATITUDE_DD)
            except Exception:
                print('not valid', pos_row.KEY)

        xs, ys = convert_projection(position_df['LATITUDE_DD'].astype(float).values,
                                    position_df['LONGITUDE_DD'].astype(float).values)
        position_df['LONGI'] = xs
        position_df['LATIT'] = ys

        comnts = []
        for key in position_df['KEY']:
            ds_meta = self.datasets[self.key_ds_mapper.get(key)]['metadata']
            cv_boolean = ds_meta.str.startswith('//METADATA;COMNT_VISIT;')
            value = ds_meta[cv_boolean].values[0].replace('//METADATA;COMNT_VISIT;', '')
            comnts.append(value)
        position_df['COMNT_VISIT'] = comnts

        self.monthly_keys = self._get_monthly_keys(position_df)
        self.position_source = ColumnDataSource(data=position_df)
        self.position_plot_source = ColumnDataSource(data=position_df)

    def _setup_month_selector(self):
        """Set month selection widget."""
        callback = cbs.month_selection_callback(position_source=self.position_source,
                                                position_plot_source=self.position_plot_source)
        self.month_selector = Select(
            title="Select month",
            value='All',
            options=['All'] + pd.date_range(start='2020-01', freq='M',
                                            periods=12).month_name().to_list()
        )
        self.month_selector.js_on_change('value', callback)
        # self.month_selector.title.text_align = 'center'
        callback.args["month"] = self.month_selector

    def _setup_download_button(self, savepath):
        """Set download widget."""
        self.download_button = cbs.get_download_widget(self.datasets,
                                                       self.position_plot_source,
                                                       self.ctd_session,
                                                       savepath)

    def _setup_get_file_button(self):
        """Set file button widget."""
        self.file_button = cbs.get_file_widget()

    def _setup_serie_table(self):
        """Create data table associated to the position_plot_source object."""
        columns = [TableColumn(field="STATION", title="Station"),
                   TableColumn(field="KEY", title="Key"),
                   ]
        self.selected_series = DataTable(source=self.position_plot_source, columns=columns,
                                         width=300, height=322)

    def _setup_flag_widgets(self):
        """Set flag widgets."""
        self.flag_widgets = {}
        for fig_key in self.figures.keys():
            if fig_key.startswith('COMBO'):
                continue
            parameter = self.plot_parameters_mapping.get(fig_key).split()[0]
            self.flag_widgets[fig_key] = cbs.get_flag_buttons_widget(
                self.position_plot_source,
                self.data_source['main_source'],
                self.datasets,
                figure_objs=self.figures,
                flag_keys=self.plot_parameters_mapping[parameter].get('q_flags'),
                color_keys=self.plot_parameters_mapping[parameter].get('color_keys'),
                select_button=self.select_all_button
            )

    def _setup_comnt_inputs(self):
        """Set comment input widgets."""
        self.comnt_visit = TextInput(value="", title="COMNT_VISIT:")
        self.comnt_visit_button = cbs.comnt_visit_change_button(
            datasets=self.datasets,
            position_source=self.position_plot_source,
            comnt_obj=self.comnt_visit,
        )

        self.comnt_samp = TextInput(value="", title="COMNT_SAMP:")
        self.comnt_samp_button = cbs.comnt_samp_change_button(
            datasets=self.datasets,
            position_source=self.position_plot_source,
            data_source=self.data_source['main_source'],
            comnt_obj=self.comnt_samp,
        )

        self.comnt_samp_selector = cbs.comnt_samp_selection(
            data_source=self.data_source['main_source'],
            comnt_obj=self.comnt_samp,
        )

    def _setup_info_block(self):
        """Set text block objects."""
        self.info_block = get_info_block()
        self.text_export = get_export_info_block()
        self.text_index_selection = standard_block_header(text='Profile index selection', height=30)
        self.text_multi_serie_flagging = standard_block_header(
            text='Multi serie parameter flagging', height=30
        )
        self.text_meta = standard_block_header(text='Change comments', height=30)
        self.text_import = standard_block_header(text='Not yet applicable', height=30)

    def _setup_selection_widgets(self):
        """Set selection widgets."""
        self.select_all_button = cbs.select_button(data_source=self.data_source['main_source'])
        self.deselect_all_button = cbs.deselect_button(data_source=self.data_source['main_source'])

        self.pressure_slider = RangeSlider(start=0, end=100, value=(0, 100),
                                           step=0.5, title="Select with pressure range", width=300)
        callback = cbs.range_selection_callback(data_source=self.data_source['main_source'])
        self.pressure_slider.js_on_change('value', callback)

    def _setup_multiflag_widget(self):
        """Set multiflag widgets."""
        def sorted_params(plist):
            para_list = []
            i = 0
            while i < len(plist):
                if '2' in plist[i]:
                    para_list.extend([plist[i + 1], plist[i]])
                    i += 2
                else:
                    para_list.append(plist[i])
                    i += 1
            return para_list

        parameter_list = []
        for p in self.figures.keys():
            if not p.startswith('COMBO'):
                parameter_list.append(self.plot_parameters_mapping.get(p).split()[0])

        parameter_list = sorted_params(sorted(parameter_list))

        self.parameter_selector = Select(title="Select parameter",
                                         value=parameter_list[0],
                                         options=parameter_list
                                         )

        self.multi_flag_widget = cbs.get_multi_serie_flag_widget(
            self.position_plot_source,
            self.data_source,
            self.datasets,
            parameter_selector=self.parameter_selector,
            parameter_mapping=self.plot_parameters_mapping,
            figure_objs=None
        )

    def _setup_reset_callback(self, **kwargs):
        """Set linked figure reset functionality.

        Autoreset all figures.
        """
        for p in self.figures.keys():
            xr_cbs = (xr_cb for xr_cb in kwargs.values())
            self.figures[p].js_on_event('reset', cbs.reset_callback(self.seconds),
                                        *xr_cbs)

    def _setup_datasource_callbacks(self):
        """Set multiflag widgets."""
        set_button_type_callback = cbs.change_button_type_callback(button=self.select_all_button,
                                                                   btype='default')
        self.data_source['main_source'].selected.js_on_change('indices', set_button_type_callback)

    def _setup_map(self):
        """Initiate the map object including all the connected interactivity."""
        pan = PanTool()
        save = SaveTool()
        tap = TapTool()
        lasso = LassoSelectTool()
        reset = ResetTool()
        wheel = WheelZoomTool()

        tooltips = HoverTool(tooltips=[("Station", "@STATION"),
                                       ("Serie", "@KEY")])

        # range bounds supplied in web mercator coordinates
        self.map = figure(x_range=(0, 4000000), y_range=(7100000, 9850000),
                          x_axis_type="mercator", y_axis_type="mercator",
                          plot_height=420, plot_width=1000,
                          tools=[pan, wheel, tap, lasso, tooltips, reset, save])

        self.map.yaxis.axis_label = ' '  # in order to aline y-axis with figure window below
        self.map.xgrid.grid_line_color = None
        self.map.ygrid.grid_line_color = None
        self.map.toolbar.active_scroll = self.map.select_one(WheelZoomTool)
        self.map.add_tile(self.tile_provider)

        station_data_callback = cbs.station_callback(position_source=self.position_plot_source,
                                                     data_source=self.data_source,
                                                     figures=self.figures,
                                                     seconds=self.seconds,
                                                     pmap=self.plot_parameters_mapping)
        tap.callback = station_data_callback

        # When we mark stations on the map using lasso selection, we activate the TS-diagram.
        lasso_callback = cbs.lasso_callback(monthly_keys=self.monthly_keys,
                                            in_data=self.ts_source,
                                            plot_data=self.ts_plot_source,
                                            x_range=self.ts.x_range,
                                            y_range=self.ts.y_range)

        station_data_callback_2 = cbs.station_callback(position_source=self.position_plot_source,
                                                       data_source=self.data_source,
                                                       figures=self.figures,
                                                       seconds=self.seconds,
                                                       pmap=self.plot_parameters_mapping)

        comnt_callback = cbs.comnt_callback(position_source=self.position_plot_source,
                                            comnt_obj=self.comnt_visit,
                                            single_select=1)

        comnt_samp_callback = cbs.comnt_samp_callback(
            position_source=self.position_plot_source,
            data_source=self.data_source['main_source'],
            comnt_obj=self.comnt_samp,
            comnt_selector=self.comnt_samp_selector,
            single_select=1,
        )

        update_slider_callback = cbs.range_slider_update_callback(
            slider=self.pressure_slider,
            data_source=self.data_source['main_source'],
        )

        select_button_type_callback = cbs.change_button_type_callback(
            button=self.select_all_button,
            btype='default'
        )

        lasso_callback.args["month"] = self.month_selector
        self.position_plot_source.selected.js_on_change(
            'indices',
            lasso_callback,
            station_data_callback_2,
            comnt_callback,
            update_slider_callback,
            select_button_type_callback,
            comnt_samp_callback,
        )

    def plot_stations(self):
        """Plot loaded stations on the map."""
        renderer = self.map.circle('LONGI', 'LATIT', source=self.position_plot_source,
                                   color="#5BC798", line_color="aquamarine", size=10, alpha=0.7)
        selected_circle = Circle(fill_alpha=0.5, fill_color="#FF0202", line_color="aquamarine")
        nonselected_circle = Circle(fill_alpha=0.3, fill_color="#5BC798", line_color="aquamarine")
        renderer.selection_glyph = selected_circle
        renderer.nonselection_glyph = nonselected_circle

    def plot_data(self):
        """Plot default data in figures.

        Default data: x=[1] and y=[1].
        """
        combo_mapping = {
            'COMBO_TEMP': ('x1', 'x4'), 'COMBO_SALT': ('x2', 'x5'), 'COMBO_DOXY': ('x3', 'x6')
        }

        nonselected_circle = Circle(fill_alpha=0.1, fill_color="#898989", line_color="lightgrey")

        for p, item in self.figures.items():
            if p.startswith('COMBO'):
                p1, p2 = combo_mapping.get(p)
                item.line(
                    p1, 'y', color="color_{}".format(p1), line_color="navy", line_width=1,
                    alpha=0.3,
                    source=self.data_source['main_source']
                )
                item.circle(
                    p1, 'y', color="color_{}".format(p1), line_color="white", size=6, alpha=0.5,
                    source=self.data_source['main_source']
                )

                item.line(
                    p2, 'y', color="color_{}".format(p2), line_color="navy", line_width=1,
                    alpha=0.3,
                    source=self.data_source['main_source']
                )
                item.cross(
                    p2, 'y', color="color_{}".format(p2), size=6, alpha=0.5,
                    source=self.data_source['main_source']
                )

                item.y_range.flipped = True
            else:
                item.line(
                    p, 'y', color="color_{}".format(p), line_color="navy", line_width=1, alpha=0.3,
                    source=self.data_source['main_source']
                )
                renderer = item.circle(
                    p, 'y', color="color_{}".format(p), line_color="white", size=6, alpha=0.5,
                    source=self.data_source['main_source']
                )
                renderer.nonselection_glyph = nonselected_circle
                item.y_range.flipped = True

        # T/S - diagram
        self.ts.circle('x', 'y', color='color', size=3, alpha=0.8,
                       source=self.ts_plot_source, legend_label='Sensor 1')
        self.ts.toolbar.active_scroll = self.ts.select_one(WheelZoomTool)
        self.ts.legend.location = "top_left"

        number_of_colors = int(self.ts_source.data[self.plot_parameters_mapping.get('y')].max()) * 2
        number_of_colors += 1
        cm_map = cm.get_cmap('cool', number_of_colors)
        color_array = [colors.to_hex(cm_map(c)) for c in range(number_of_colors)]
        color_bar = ColorBar(
            color_mapper=LinearColorMapper(
                palette=color_array,
                low=0,
                high=self.ts_source.data[self.plot_parameters_mapping.get('y')].max()
            ),
            location=(0, 0),
        )

        self.ts.add_layout(color_bar, 'right')

        x_min, x_max, y_min, y_max = 0, 40, -10, 30
        contour_data = get_contour_data(x_min, x_max, y_min, y_max)
        for key in contour_data.keys():
            self.ts.line(contour_data[key]['salt'], contour_data[key]['temp'],
                         line_color="grey", line_alpha=0.8, line_width=1.5)

    def append_qc_comment(self, meta_series):
        """Append commnet at the end of the metadata serie.

        Will be placed on the row just above the data table in each data file.
        """
        time_stamp = get_time_as_format(now=True, fmt='%Y%m%d%H%M')
        meta_series[len(meta_series) + 1] = '//QC_COMNT; MANUAL QC PERFORMED BY {}; TIMESTAMP {}' \
                                            ''.format(self.ctd_session.settings.user, time_stamp)

    def get_tab_layout(self):
        """Return bokeh tab widget."""
        fig_tabs = [Panel(child=column([Spacer(height=30, width=20), self.ts]), title="TS")]
        for p, item in self.figures.items():
            if (self.multi_sensors and p not in ['x1', 'x2', 'x3', 'x4', 'x5', 'x6',
                                                 'COMBO_TEMP', 'COMBO_SALT', 'COMBO_DOXY']) \
                    or (not self.multi_sensors and p not in ['x1', 'x2', 'x3']):
                tab_layout = column([self.flag_widgets[p], item])
                tab_name = self.plot_parameters_mapping.get(p).split()[0].replace('_CTD', '')
                pan = Panel(child=tab_layout, title=tab_name)
                fig_tabs.append(pan)

        return Tabs(tabs=fig_tabs)

    def get_tabs(self, **kwargs):
        """Return bokeh tab widget."""
        tabs = []
        for name, item in kwargs.items():
            tab = self.get_column(item)
            pan = Panel(child=tab, title=name)
            tabs.append(pan)

        return Tabs(tabs=tabs)

    def get_column(self, item):
        """Return bokeh column layout."""
        c_list = []
        for attr in item:
            if type(attr) == str:
                c_list.append(self.__getattribute__(attr))
            elif type(attr) == tuple:
                r = row([self.__getattribute__(a) for a in attr], sizing_mode="stretch_width")
                c_list.append(r)
        return column(c_list)

    def get_std_parameter_tab_layout(self):
        """Return list of bokeh column layouts."""
        def pan_title(string):
            return string.split()[0].replace('_CTD', '')

        columns = []

        if self.multi_sensors:
            for params in zip(('x1', 'x2', 'x3'), ('x4', 'x5', 'x6'),
                              ('COMBO_TEMP', 'COMBO_SALT', 'COMBO_DOXY')):
                pans = []
                for p in params:
                    if p in self.figures:
                        tab_cols = []
                        if p in self.flag_widgets:
                            tab_cols.append(self.flag_widgets[p])
                        else:
                            tab_cols.append(Spacer(width=20, height=41))
                        tab_cols.append(self.figures[p])
                        tab = column(tab_cols)
                        pan = Panel(
                            child=tab, title=pan_title(self.plot_parameters_mapping.get(p) or p)
                        )
                        pans.append(pan)
                columns.append(column([Tabs(tabs=pans)]))
        else:
            for p1 in ('x1', 'x2', 'x3'):
                columns.append(column([self.flag_widgets[p1], self.figures[p1]]))

        return columns

    def get_layout(self):
        """Return the complete bokeh layout."""
        tabs = self.get_tab_layout()
        meta_tabs = self.get_tabs(Data=['text_index_selection',
                                        ('select_all_button', 'deselect_all_button'),
                                        'pressure_slider',
                                        'text_multi_serie_flagging',
                                        'parameter_selector',
                                        'multi_flag_widget'],
                                  Metadata=['text_meta',
                                            'comnt_visit', 'comnt_visit_button',
                                            'comnt_samp',
                                            'comnt_samp_selector',
                                            'comnt_samp_button'],
                                  Import=['text_import',
                                          'file_button'],
                                  Export=['text_export',
                                          'download_button'],
                                  Info=['info_block'])
        std_parameter_tabs = self.get_std_parameter_tab_layout()
        widgets_1 = column([self.month_selector, self.spacer, self.selected_series],
                           sizing_mode="fixed", height=400, width=200)
        widgets_2 = column([Spacer(height=10, width=125)],
                           sizing_mode="fixed", height=10, width=125)
        widgets_3 = column([meta_tabs], sizing_mode="stretch_both", height=100, width=100)
        return grid(
            [
                row([self.map, widgets_1, widgets_2, widgets_3]),
                row([*std_parameter_tabs, column([tabs])])
            ]
        )

    @property
    def spacer(self):
        """Return bokeh spacer."""
        return Spacer(height=10, width=20)

    def return_layout(self):
        """Return the layout.

        Can be displayed in an embedded bokeh server.
        """
        return self.get_layout()

    def show_plot(self):
        """Show layout in a html-file or in jupyter notebook."""
        show(self.get_layout())



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:42

@author: a002028
"""
import json
from ctdvis import utils


def load_json(file_path):
    """Return json data."""
    with open(file_path, 'r', encoding='utf-8') as fd:
        f = json.load(fd)
    return f


class JSONreader(dict):
    """Handler of json reader/writer.

    - Import json
    - Export to json
    - Find dictionary within json file based on a specific key
    - Add elements to dictionary
    - Fill up json/dictionary structure with relevant/desired information
    """

    # TODO: Divide into proper reader and writer!

    def _export_json(self, data_dict=None, out_source='', indent=4):
        """Write data to json file."""
        data_dict = data_dict or {}
        with open(out_source, "w") as outfile:
            json.dump(data_dict, outfile, indent=indent)

    def _initiate_attributes(self):
        """Doc."""
        pass

    def _initiate_outfile(self):
        """Initiate out file list.

        Save multiple dictionaries stored in a list.
        """
        self.out_file = []

    def _get_dictionary_reference(self, dictionary=None, dict_path=None):
        """Return dict."""
        dictionary = dictionary or {}
        dict_path = dict_path or []
        for key in dict_path:
            if isinstance(key, str) and key not in dictionary:
                return None
            dictionary = dictionary[key]
        return dictionary

    def export(self, out_source='', out_file=None):
        """Guide export method."""
        if out_file:
            self._export_json(out_source=out_source,
                              data_dict=out_file)

        elif hasattr(self, 'out_file'):
            self._export_json(out_source=out_source,
                              data_dict=self.out_file)

        elif hasattr(self, 'config'):
            self._export_json(out_source=out_source,
                              data_dict=self.config)

        else:
            raise UserWarning('No outfile specified for export to .json')

    def load_json(self, config_files=None, return_dict=False):
        """Load json data file(s)."""
        if isinstance(config_files, str):
            config_files = [config_files]

        for config_file in config_files:
            with open(config_file, 'r', encoding='utf-8') as fd:
                self = utils.recursive_dict_update(self, json.load(fd))

        if return_dict:
            return self



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-12-15 13:59

@author: johannes
"""
import yaml


def yaml_reader(file_path):
    """Return data from yaml file."""
    with open(file_path, encoding='utf8') as fd:
        data = yaml.load(fd, Loader=yaml.Loader)
    return data



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-06 15:33

@author: a002028
"""
from functools import partial
from bokeh.models import Button, FileInput, CustomJS, CrosshairTool
from bokeh.layouts import row, Spacer
from bokeh.models.widgets import Select
from bokeh.plotting import figure
from bokeh.events import ButtonClick
from ctdvis.utils import get_time_as_format


def callback_test(source):
    """Return a CustomJS callback.

    Intended for test use. printing 'console print callback_test!'.
    """
    code = """
    // CALLBACK TESTING, WITH PRINT
    console.log('console print callback_test!')
    """
    # Create a CustomJS callback
    return CustomJS(args={'source': source},
                    code=code)


def month_selection_callback(position_source=None, position_plot_source=None):
    """Return a CustomJS callback."""
    # TODO We need to rewrite the JScallback code.. not very versatile..
    #  simply use all columns from position_source to begin with?
    code = """
    console.log('month_selection_callback');
    // Get data from ColumnDataSource
    var selected_data = {LATIT: [], LONGI: [], STATION: [], KEY: [], MONTH: [], COMNT_VISIT: []};
    var data = source.data;

    var month_mapping = {'All': 'All',
                         'January': '01', 'February': '02',
                         'March': '03', 'April': '04',
                         'May': '05', 'June': '06',
                         'July': '07', 'August': '08',
                         'September': '09', 'October': '10',
                         'November': '11', 'December': '12'};

    var selected_month = month_mapping[month.value];

    var key_val, lat_val, lon_val, statn_val, mon_val, cvis_val;
    for (var i = 0; i < data.KEY.length; i++) {
        key_val = data.KEY[i];
        lat_val = data.LATIT[i];
        lon_val = data.LONGI[i];
        statn_val = data.STATION[i];
        mon_val = data.MONTH[i];
        cvis_val = data.COMNT_VISIT[i];

        if (selected_month == 'All') {
            selected_data.KEY.push(key_val);
            selected_data.LATIT.push(lat_val);
            selected_data.LONGI.push(lon_val);
            selected_data.STATION.push(statn_val);
            selected_data.MONTH.push(mon_val);
            selected_data.COMNT_VISIT.push(cvis_val);
        } else if (mon_val == selected_month) {
            selected_data.KEY.push(key_val);
            selected_data.LATIT.push(lat_val);
            selected_data.LONGI.push(lon_val);
            selected_data.STATION.push(statn_val);
            selected_data.MONTH.push(mon_val);
            selected_data.COMNT_VISIT.push(cvis_val);
        }
    }

    plot_source.data = selected_data;
    """
    # Create a CustomJS callback.
    return CustomJS(args={'source': position_source,
                          'plot_source': position_plot_source},
                    code=code)


def station_callback(position_source=None, data_source=None,
                     figures=None, seconds=None, pmap=None):
    """Return a CustomJS callback."""
    # assert position_source, data_source
    code = """
    //console.log('station_callback');
    // Set column name to select similar glyphs
    var key = 'KEY';
    var statn_key = 'STATION';
    var sec = seconds.data;

    // Get data from ColumnDataSource
    var position_data = position_source.data;
    //var data = data_source.data;
    var parameter_mapping = parameter_mapping;
    var figures = figures;

    //console.log('parameter_mapping', parameter_mapping);

    // Get indices array of all selected items
    var selected = position_source.selected.indices;

    //console.log('data[y].length', data['y'].length)
    //console.log('selected', selected);

    // Update figure titles flag_color_mapping
    var station_name = position_data[statn_key][selected[0]];
    var selected_key = position_data[key][selected[0]];

    //console.log('station_name', station_name);
    //console.log('selected_key', selected_key);

    // Update active keys in data source
    if (selected.length == 1) {
        for (var fig_key in figures){
            figures[fig_key].title.text = station_name + ' - ' + selected_key
        }
        data_source['main_source'].data = data_source[selected_key].data;

        // Save changes to ColumnDataSource
        data_source['main_source'].change.emit();
    } else {
        data_source['main_source'].data = data_source['default_source'].data;
        console.log('We can only work with one serie at a time', selected.length)
    }

    var d = new Date();
    var t = d.getTime();
    var new_seconds = Math.round(t / 1000);
    sec.tap_time[0] = new_seconds;
    sec.reset_time[0] = new_seconds;
    seconds.change.emit();
    for (var fig_key in figures){
        figures[fig_key].reset.emit();
    }
    //console.log('station_callback - DONE');
    """
    # Create a CustomJS callback with the code and the data
    return CustomJS(args={'position_source': position_source,
                          'data_source': data_source,
                          'figures': figures,
                          'seconds': seconds,
                          'parameter_mapping': pmap,
                          },
                    code=code)


def lasso_callback(monthly_keys=None, in_data=None, plot_data=None, x_range=None, y_range=None):
    """Return a CustomJS callback."""
    code = """
    //console.log('lasso_callback');
    var month_mapping = {'All': 'All',
                         'January': '01', 'February': '02',
                         'March': '03', 'April': '04',
                         'May': '05', 'June': '06',
                         'July': '07', 'August': '08',
                         'September': '09', 'October': '10',
                         'November': '11', 'December': '12'};

    var selected_month = month_mapping[month.value];

    var data = {x: [], y: [], color: [], key: []};
    var indices = cb_obj.indices;
    var selected_keys = [];
    for (var i = 0; i < indices.length; i++) {
        selected_keys.push(monthly_keys[selected_month][indices[i]]);
    }

    //console.log('selected_keys', selected_keys)

    var key_val, x_val, y_val, c_val;
    for (var i = 0; i < in_data.KEY.length; i++) {
        key_val = in_data.KEY[i];
        x_val = in_data.x[i];
        y_val = in_data.y[i];
        c_val = in_data.color[i];

        if (selected_keys.indexOf(key_val) !== -1) {
            data.x.push(x_val);
            data.y.push(y_val);
            data.color.push(c_val);
            data.key.push(key_val);
        }
    }
    if (data.x.length > 1) {
        //console.log('Update!')
        plot_data.data = data;
        x_range.start = Math.min.apply(Math, data.x)-0.5;
        x_range.end = Math.max.apply(Math, data.x)+0.5;
        y_range.start = Math.min.apply(Math, data.y)-0.5;
        y_range.end = Math.max.apply(Math, data.y)+0.5;
        x_range.change.emit();
        y_range.change.emit();
        //console.log('x_range.start', x_range.start)
    }
    """
    return CustomJS(args=dict(monthly_keys=monthly_keys,
                              in_data=in_data.data,
                              plot_data=plot_data,
                              x_range=x_range,
                              y_range=y_range),
                    code=code)


def comnt_callback(position_source=None, comnt_obj=None, single_select=None):
    """Return a CustomJS callback."""
    code = """
    //console.log('comnt_callback');
    // Set column name to select similar glyphs
    var key = 'KEY';
    var statn_key = 'STATION';
    var comnt_key = 'COMNT_VISIT';

    // Set Sources
    var position_data = position_source.data;
    var comnt_obj = comnt_obj;
    var title_obj = title_obj;
    var single_select = single_select;

    // Get indices array of all selected items
    var selected = position_source.selected.indices;

    // Update figure title
    var station_name = position_data[statn_key][selected[0]];
    var selected_key = position_data[key][selected[0]];
    var comnt = position_data[comnt_key][selected[0]];

    // Update active keys in data source
    if ((single_select == 1 && selected.length == 1) || (single_select == 0)) {
        comnt_obj.value = comnt;
        comnt_obj.title = 'COMNT_VISIT: ' + station_name + ' - ' + selected_key;
    }

    """
    # Create a CustomJS callback with the code and the data
    return CustomJS(args={'position_source': position_source,
                          'comnt_obj': comnt_obj,
                          'single_select': single_select,
                          },
                    code=code)


def comnt_samp_callback(position_source=None, comnt_obj=None, data_source=None,
                        comnt_selector=None, single_select=None):
    """Return a CustomJS callback."""
    code = """
    // Set column name to select similar glyphs
    var key = 'KEY';
    var statn_key = 'STATION';

    // Set Sources
    var position_data = position_source.data;
    var data = data_source.data;
    var comnt_obj = comnt_obj;
    var single_select = single_select;

    // Get indices array of all selected items
    var selected = position_source.selected.indices;

    // Update active keys in data source
    if ((single_select == 1 && selected.length == 1) || (single_select == 0)) {
        // Update figure title
        var station_name = position_data[statn_key][selected[0]];
        var selected_key = position_data[key][selected[0]];
        comnt_obj.title = 'COMNT_SAMP:  ' + station_name + ' - ' + selected_key;
        comnt_obj.value = '';
        function onlyUnique(value, index, self) {
          return self.indexOf(value) === index;
        }
        var unique_values = data['COMNT_SAMP'].filter(onlyUnique);
        unique_values = unique_values.filter(function(item) {
            return item !== ""
        })
        var using = [''];
        using.push(...unique_values);
        comnt_selector.options = using;
    }

    """
    return CustomJS(
        args={
            'position_source': position_source,
            'comnt_obj': comnt_obj,
            'comnt_selector': comnt_selector,
            'data_source': data_source,
            'single_select': single_select,
        },
        code=code
    )


def change_button_type_callback(button=None, btype=None):
    """Return a CustomJS callback."""
    code = """
    button.button_type = btype;
    """
    return CustomJS(args={'button': button, 'btype': btype}, code=code)


def select_button(data_source=None):
    """Return a button."""
    code = """
    var data = data_source.data;
    var indices = [];

    var i = 0;
    while ( ! isNaN(data.y[i]) ) {
        indices.push(i)
        i++
    }
    data_source.selected.indices = indices;
    button.button_type = 'success';
    //console.log('select_button DONE');
    """
    button = Button(label="Select all indices", width=30, button_type="default")
    callback = CustomJS(args={'data_source': data_source, 'button': button}, code=code)
    button_type_callback = change_button_type_callback(button=button, btype='success')
    button.js_on_event(ButtonClick, callback, button_type_callback)
    return button


def deselect_button(data_source=None):
    """Return a button."""
    code = """
    data_source.selected.indices = [];
    """
    callback = CustomJS(args={'data_source': data_source}, code=code)
    button = Button(label="Deselect all indices", width=30, button_type="default")
    button.js_on_event(ButtonClick, callback)
    return button


def range_slider_update_callback(slider=None, data_source=None):
    """Return a CustomJS callback."""
    code = """
    var data = data_source.data;
    var values = [];
    var i = 0;
    while ( ! isNaN(data.y[i]) ) {
        values.push(data.y[i])
        i++
    }
    slider.start = Math.min.apply(Math, values);
    if (slider.start !== Math.max.apply(Math, values)) {
            slider.end = Math.max.apply(Math, values);
    } else {
        slider.end = Math.max.apply(Math, values) + 1;
    }
    slider.value = [slider.start, slider.end];
    slider.change.emit();
    """
    return CustomJS(args={'slider': slider, 'data_source': data_source},
                    code=code)


def range_selection_callback(data_source=None):
    """Return a CustomJS callback."""
    code = """
    var data = data_source.data;
    var min_pres = cb_obj.value[0];
    var max_pres = cb_obj.value[1];
    var indices = [];
    for (var i = 0; i < data.y.length; i++) {
        if ((data.y[i] >= min_pres) && (data.y[i] <= max_pres)) {
            indices.push(i)
        }
    }
    data_source.selected.indices = indices;
    """
    return CustomJS(args={'data_source': data_source},
                    code=code)


def get_flag_widget(position_source, data_source, flag_key=None, color_key=None):
    """Return a bokeh row layout.

    Flag selector.
    """
    code = """
    console.log('get_flag_widget');
    var flag_color_mapping = {'A-flag': {'c':'navy', 'flag': ''},
                              'B-flag': {'c':'red', 'flag': 'B'},
                              'E-flag': {'c':'green', 'flag': 'E'},
                              'S-flag': {'c':'orange', 'flag': 'S'}};

    // Get data from ColumnDataSource
    var position_data = position_source.data;
    var data = data_source.data;

    // Set variables attributes
    var color_column = color_key;
    var selected_flag = flag_selection.value;

    var selected_position = position_source.selected.indices;
    var selected_key = position_data['KEY'][selected_position[0]];
    var flag_column = selected_key+'_'+flag_key;

    // Get indices array of all selected items
    var selected_indices = data_source.selected.indices;

    for (var i = 0; i < selected_indices.length; i++) {
        data[color_column][selected_indices[i]] = flag_color_mapping[selected_flag]['c'];
        data[flag_column][selected_indices[i]] = flag_color_mapping[selected_flag]['flag'];
        // console.log('data[flag_column][selected_indices[i]]', data[flag_column][selected_indices[i]])
    }

    // Save changes to ColumnDataSource
    data_source.change.emit();
    """  # noqa: E501
    callback = CustomJS(args={'position_source': position_source,
                              'data_source': data_source},
                        code=code)
    flag_selector = Select(value='A-flag',
                           options=['A-flag', 'B-flag', 'E-flag', 'S-flag'],
                           width=100, height=50)
    callback.args["flag_selection"] = flag_selector
    callback.args["color_key"] = color_key
    callback.args["flag_key"] = flag_key
    button = Button(label="Flag Data", width=50)
    button.js_on_event(ButtonClick, callback)

    return row([flag_selector, button], sizing_mode="stretch_width")


def get_flag_buttons_widget(position_source, data_source, datasets, flag_keys=None,
                            color_keys=None, figure_objs=None, select_button=None):
    """Return a list of buttons.

    Each button represents a QC-flag which will be applied when the button is pressed.
    """
    code = """
    //console.log('get_flag_buttons_widget');
    var flag_color_mapping = {'A-flag': {'c':'navy', 'flag': ''},
                              'B-flag': {'c':'red', 'flag': 'B'},
                              'E-flag': {'c':'green', 'flag': 'E'},
                              'S-flag': {'c':'orange', 'flag': 'S'}};

    // Get data from ColumnDataSource
    var position_data = position_source.data;
    var data = data_source.data;
    var select_button_type = select_button.button_type;

    // Set variables attributes
    var color_columns = color_keys;
    var flag_keys = flag_keys;
    var selected_flag = flag;

    var selected_position = position_source.selected.indices;
    var selected_key = position_data['KEY'][selected_position[0]];

    // Get indices array of all selected items
    var selected_indices = data_source.selected.indices;

    var flag_value = flag_color_mapping[selected_flag]['flag'];
    var color_value = flag_color_mapping[selected_flag]['c'];

    if (selected_position.length == 1) {
        for (var i = 0; i < selected_indices.length; i++) {
            //console.log('selected_indices[i]', selected_indices[i])
            //console.log('index_value', index_value)
            for (var j = 0; j < color_columns.length; j++) {
                data[color_columns[j]][selected_indices[i]] = color_value;
                //data[flag_keys[j]][selected_indices[i]] = flag_value;
            }
        }

        // Save changes to ColumnDataSource (only on the plotting side of ColumnDataSource)
        data_source.change.emit();
        for (var key in figure_objs) {
            figure_objs[key].reset.emit();
        }
        data_source.selected.indices = selected_indices;
        select_button.button_type = select_button_type;

        // Trigger python callback inorder to save changes to the actual datasets
        dummy_trigger.glyph.size = {'value': Math.random(), 'units': 'screen'};
        dummy_trigger.glyph.change.emit();

    } else {
        console.log('To many selected stations!! We can only work with one at a time', selected_position.length)
    }
    """  # noqa: E501
    flag_color_mapping = {'A-flag': {'c': 'navy', 'flag': ''},
                          'B-flag': {'c': 'red', 'flag': 'B'},
                          'E-flag': {'c': 'green', 'flag': 'E'},
                          'S-flag': {'c': 'orange', 'flag': 'S'}}

    def callback_py(attr, old, new, flag=None):
        selected_position = position_source.selected.indices
        if len(selected_position) > 1:
            print('multi serie selection, no good! len(selected_position) = {}'
                  ''.format(len(selected_position)))
            return

        selected_key = position_source.data['KEY'][selected_position[0]]
        selected_indices = data_source.selected.indices
        ds_key = ''.join(('ctd_profile_', selected_key, '.txt'))
        flag_value = flag_color_mapping[flag].get('flag')
        datasets[ds_key]['data'].loc[selected_indices, flag_keys] = flag_value

    # button_types = default, primary, success, warning or danger
    button_types = ['primary', 'danger', 'success', 'warning']
    flag_list = ['A-flag', 'B-flag', 'E-flag', 'S-flag']
    button_list = [Spacer(width=10, height=10)]
    dummy_figure = figure()
    for flag, b_type in zip(flag_list, button_types):
        dummy_trigger = dummy_figure.circle(x=[1], y=[1], alpha=0)
        dummy_trigger.glyph.on_change('size', partial(callback_py, flag=flag))

        callback = CustomJS(args={'position_source': position_source,
                                  'data_source': data_source,
                                  'figure_objs': figure_objs,
                                  'flag': flag,
                                  'dummy_trigger': dummy_trigger,
                                  'select_button': select_button},
                            code=code)

        callback.args["color_keys"] = color_keys
        callback.args["flag_keys"] = flag_keys

        button = Button(label=flag, width=30, button_type=b_type)
        button.js_on_event(ButtonClick, callback)

        button_list.append(button)

    button_list.append(Spacer(width=10, height=10))

    return row(button_list, sizing_mode="stretch_width")


def get_multi_serie_flag_widget(position_source, data_source, datasets, parameter_selector=None,
                                parameter_mapping=None, figure_objs=None):
    """Return a list of buttons.

    Each button represents a QC-flag which will be applied when the button is pressed.
    """
    code = """
    console.log('get_multi_serie_flag_widget');
    var flag_color_mapping = {'A-flag': {'c':'navy', 'flag': ''},
                              'B-flag': {'c':'red', 'flag': 'B'},
                              'E-flag': {'c':'green', 'flag': 'E'},
                              'S-flag': {'c':'orange', 'flag': 'S'}};

    // Get data from ColumnDataSource
    var position_data = position_source.data;
    // var data = data_source.data;

    // Set variables attributes
    var selected_flag = flag;

    var flag_keys = parameter_mapping[parameter_selector.value]['q_flags']
    var color_columns = parameter_mapping[parameter_selector.value]['color_keys']

    var flag_value = flag_color_mapping[selected_flag]['flag'];
    var color_value = flag_color_mapping[selected_flag]['c'];

    var selected_position_indices = position_source.selected.indices;
    var selected_key = 0;
    var value_array = [];
    var valid_indices = [];

    for (var i_pos = 0; i_pos < selected_position_indices.length; i_pos++) {
        var selected_key = position_data['KEY'][selected_position_indices[i_pos]];
        var value_array = data_source[selected_key].data['x1'];

        for (var i = 0; i < value_array.length; i++) {
            for (var j = 0; j < color_columns.length; j++) {
                data_source[selected_key].data[color_columns[j]][i] = color_value;
            }
        }
        data_source[selected_key].change.emit();
    }

    for (var key in figure_objs) {
        figure_objs[key].reset.emit();
    }

    // Trigger python callback inorder to save changes to the actual datasets
    dummy_trigger.glyph.size = {'value': Math.random(), 'units': 'screen'};
    dummy_trigger.glyph.change.emit();
    console.log('DONE - get_multi_serie_flag_widget');
    """
    flag_color_mapping = {'A-flag': {'c': 'navy', 'flag': ''},
                          'B-flag': {'c': 'red', 'flag': 'B'},
                          'E-flag': {'c': 'green', 'flag': 'E'},
                          'S-flag': {'c': 'orange', 'flag': 'S'}}

    def callback_py(attr, old, new, flag=None):
        flag_keys = parameter_mapping[parameter_selector.value].get('q_flags')

        selected_position = position_source.selected.indices
        for pos_source_index in selected_position:
            selected_key = position_source.data['KEY'][pos_source_index]
            ds_key = ''.join(('ctd_profile_', selected_key, '.txt'))
            flag_value = flag_color_mapping[flag].get('flag')
            datasets[ds_key]['data'].loc[:, flag_keys] = flag_value

    # button_types = default, primary, success, warning or danger
    button_types = ['primary', 'danger', 'success', 'warning']
    flag_list = ['A-flag', 'B-flag', 'E-flag', 'S-flag']
    button_list = []
    dummy_figure = figure()
    for flag, b_type in zip(flag_list, button_types):
        dummy_trigger = dummy_figure.circle(x=[1], y=[2], alpha=0)
        dummy_trigger.glyph.on_change('size', partial(callback_py, flag=flag))

        callback = CustomJS(args={'position_source': position_source,
                                  'data_source': data_source,
                                  'figure_objs': figure_objs,
                                  'parameter_mapping': parameter_mapping,
                                  'parameter_selector': parameter_selector,
                                  'flag': flag,
                                  'dummy_trigger': dummy_trigger},
                            code=code)

        button = Button(label=flag, width=30, button_type=b_type)
        button.js_on_event(ButtonClick, callback)

        button_list.append(button)

    return row(button_list, sizing_mode="stretch_width")


def get_download_widget(datasets, series, session, savepath):
    """Return a download button."""
    def callback_download(event):
        def serie_generator(datasets_filelist, selected_keylist):
            for name in datasets_filelist:
                for key in selected_keylist:
                    if key in name:
                        yield name, key

        def append_qc_comment(meta):
            time_stamp = get_time_as_format(now=True, fmt='%Y%m%d%H%M')
            meta[len(meta) + 1] = '//COMNT_QC; MANUAL QC PERFORMED BY {}; TIMESTAMP {}'.format(
                session.settings.user, time_stamp)

        if not any(series.selected.indices):
            print('No selected series to download')
            print('len(series.selected.indices)', series.selected.indices)
            return

        generator = serie_generator(
            datasets.keys(),
            [series.data['KEY'][idx] for idx in series.selected.indices]
        )

        datasets_to_update = {}
        for ds_name, _ in generator:
            append_qc_comment(datasets[ds_name]['metadata'])
            datasets_to_update[ds_name] = datasets[ds_name]

        if any(datasets_to_update):
            session.save_data(
                [datasets_to_update],
                save_path=savepath or 'C:/QC_CTD',
                writer='ctd_standard_template',
            )
        else:
            print('No download!')

    button = Button(label="Download selected data", button_type="success", width=40)
    button.on_event(ButtonClick, callback_download)
    return button


def comnt_visit_change_button(datasets=None, position_source=None, comnt_obj=None):
    """Return a button."""
    def callback_py(attr, old, new, comnt_obj=None):
        selected_indices = position_source.selected.indices
        if len(selected_indices) > 1:
            print('multi serie selection, no good! len(selected_position) = {}'
                  ''.format(len(selected_indices)))
            return
        selected_key = position_source.data['KEY'][selected_indices[0]]
        ds_key = ''.join(('ctd_profile_', selected_key, '.txt'))
        cv_boolean = datasets[ds_key]['metadata'].str.startswith('//METADATA;COMNT_VISIT;')
        datasets[ds_key]['metadata'][cv_boolean] = '//METADATA;COMNT_VISIT;' + comnt_obj.value

    js_code = """
    console.log('comnt_visit_change_button')
    // Get data from ColumnDataSource
    var comnt_column = 'COMNT_VISIT';
    var position_data = position_source.data;

    // Set variables attributes
    var new_comnt = comnt_obj.value;
    var selected_indices = position_source.selected.indices;

    if (selected_indices.length == 1) {
        position_data[comnt_column][selected_indices[0]] = new_comnt;

        // Save changes to ColumnDataSource
        position_source.change.emit();

        // Trigger python callback inorder to save changes to the actual datasets
        dummy_trigger.glyph.size = {'value': Math.random(), 'units': 'screen'};
        dummy_trigger.glyph.change.emit();

    } else {
        console.log('To many selected stations!! We can only work with one at a time', selected_indices.length)
    }
    """  # noqa: E501
    dummy_figure = figure()
    dummy_trigger = dummy_figure.circle(x=[1], y=[2], alpha=0)
    dummy_trigger.glyph.on_change('size', partial(callback_py, comnt_obj=comnt_obj))

    callback = CustomJS(args={'position_source': position_source,
                              'comnt_obj': comnt_obj,
                              'dummy_trigger': dummy_trigger},
                        code=js_code)

    button = Button(label="Commit visit comnt", width=30, button_type="success")
    button.js_on_event(ButtonClick, callback)
    return button


def comnt_samp_change_button(datasets=None, position_source=None, data_source=None, comnt_obj=None):
    """Return a button."""
    def callback_py(attr, old, new, comnt_obj=None):
        selected_indices = position_source.selected.indices
        if len(selected_indices) > 1:
            print('multi serie selection, no good! len(selected_position) = {}'
                  ''.format(len(selected_indices)))
            return
        selected_key = position_source.data['KEY'][selected_indices[0]]
        selected_data_indices = data_source.selected.indices
        # data_source[selected_key].data['COMNT_SAMP'][selected_data_indices] = comnt_obj.value
        ds_key = ''.join(('ctd_profile_', selected_key, '.txt'))
        # TODO add '; '.join(...)
        # datasets[ds_key]['data']['COMNT_SAMP'].iloc[selected_data_indices] = \
        #     datasets[ds_key]['data']['COMNT_SAMP'].iloc[selected_data_indices].apply(
        #         lambda x: '; '.join((x, comnt_obj.value))
        #     )
        datasets[ds_key]['data']['COMNT_SAMP'].iloc[selected_data_indices] = comnt_obj.value
        # datasets[ds_key]['data'][cv_boolean] = comnt_obj.value

    js_code = """
    console.log('comnt_visit_change_button')
    // Get data from ColumnDataSource
    var position_data = position_source.data;
    var data = data_source.data;

    // Set variables attributes
    var selected_position_indices = position_source.selected.indices;

    if (selected_position_indices.length == 1) {
        var selected_water_indices = data_source.selected.indices;
        var selected_key = position_data['KEY'][selected_position_indices[0]];

        for (var i = 0; i < selected_water_indices.length; i++) {
            data['COMNT_SAMP'][selected_water_indices[i]] = comnt_obj.value;
        }

        // Save changes to ColumnDataSource
        data_source.change.emit();

        // Trigger python callback inorder to save changes to the actual datasets
        dummy_trigger.glyph.size = {'value': Math.random(), 'units': 'screen'};
        dummy_trigger.glyph.change.emit();

    } else {
        console.log('To many selected stations!! We can only work with one at a time', selected_indices.length)
    }
    """  # noqa: E501
    dummy_figure = figure()
    dummy_trigger = dummy_figure.circle(x=[1], y=[2], alpha=0)
    dummy_trigger.glyph.on_change('size', partial(callback_py, comnt_obj=comnt_obj))

    callback = CustomJS(args={'position_source': position_source,
                              'data_source': data_source,
                              'comnt_obj': comnt_obj,
                              'dummy_trigger': dummy_trigger},
                        code=js_code)

    button = Button(label="Commit sample comnt", width=30, button_type="success")
    button.js_on_event(ButtonClick, callback)
    return button


def comnt_samp_selection(data_source=None, comnt_obj=None):
    """Return a select widget."""
    code = """
    //console.log('select: value=' + this.value, this.toString())
    comnt_obj.value = this.value;
    var data = data_source.data;
    var selected_indices = [];
    var comnt_value;
    for (var i = 0; i < data.COMNT_SAMP.length; i++) {
        comnt_value = data.COMNT_SAMP[i];
        if (comnt_value == this.value) {
            selected_indices.push(i);
        }
    }
    data_source.selected.indices = selected_indices;
    """
    callback = CustomJS(
        code=code,
        args={'data_source': data_source,
              'comnt_obj': comnt_obj}
    )
    select = Select(
        title="Select comnt incl. index",
        value='None',
        options=[],
    )
    select.js_on_change("value", callback)
    return select


def get_file_widget():
    """Return a file input button."""
    # button_input = FileInput(accept=".csv,.txt")
    button_input = FileInput()

    return button_input


def add_hlinked_crosshairs(*figs):
    """Link crosshair between figures."""
    cht = CrosshairTool(line_alpha=0.5, dimensions="width")
    for f in figs:
        f.add_tools(cht)


def x_range_callback(x_range_obj=None, delta=4, seconds=None):
    """Return a CustomJS callback."""
    code = """
    //console.log('x_range_callback');
    var sec = seconds.data;
    var d = new Date();
    var t = d.getTime();
    var new_seconds = t / 1000;
    var tap_time_delta = new_seconds - sec.tap_time[0];
    var reset_time_delta = new_seconds - sec.reset_time[0];
    var delta_add;
    if (tap_time_delta < 1.5 || reset_time_delta < 1.5) {
        var start = xr.start;
        var end = xr.end;
        var x_delta = end - start;
        if (x_delta < accepted_delta) {
            delta_add = (accepted_delta - x_delta) / 2;
            end = end + delta_add;
            start = start - delta_add;
        }
        xr.end = end;
        xr.start = start;
    }
    """
    return CustomJS(args={'xr': x_range_obj,
                          'seconds': seconds,
                          'accepted_delta': delta},
                    code=code)


def reset_callback(seconds):
    """Return a CustomJS callback."""
    code = """
    //console.log('reset_callback');
    var sec = seconds.data;
    var d = new Date();
    var t = d.getTime();
    sec.reset_time[0] = t / 1000;
    //console.log('sec.reset_time[0]', sec.reset_time[0])
    seconds.change.emit();
    """
    return CustomJS(args={'seconds': seconds},
                    code=code)


def reset_all_callback(figures):
    """Return a CustomJS callback."""
    code = """
    console.log('reset_all_callback');
    for (var i = 0; i < figure_objs.length; i++) {
        figure_objs[i].reset.emit();
    }
    """
    return CustomJS(args={'figure_objs': figures},
                    code=code)



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-09-07 11:38

@author: a002028
"""
from bokeh.models import ColumnDataSource
from ctdvis.utils import get_color_palette


class Source(ColumnDataSource):
    """Inherit from bokeh.models.ColumnDataSource.

    Used to create a TS-datasource.
    """

    def setup_source(self, df, pmap):
        """Update data source.

        Args:
            df: pandas.DataFrame
            pmap: Dictionary with parameter mapping
        """
        parameters = [pmap.get(p) for p in ('y', 'x1', 'x2')]
        parameters.append('KEY')
        ts_df = df.loc[:, parameters].copy()
        ts_df.loc[:, 'x'] = ts_df.loc[:, pmap.get('x2')]  # x2 = SALT
        ts_df.loc[:, 'y'] = ts_df.loc[:, pmap.get('x1')]  # x1 = TEMP
        ts_df.loc[:, 'color'] = get_color_palette(dep_serie=ts_df.loc[:, pmap.get('y')])

        self.data.update(ts_df)



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:24

@author: a002028
"""
from ctdvis.config import Settings
from ctdvis.tools.quality_control import QCWorkTool
from ctdvis.datahandler import DataHandler


class Session:
    """Main class of ctdvis."""

    def __init__(self, visualize_setting=None, data_directory=None, filters=None):
        """Initiate."""
        self.data_directory = data_directory
        self.settings = Settings(visualize_setting=visualize_setting)

        self.dh = DataHandler(filters)

    def setup_datahandler(self):
        """Load data and set up the dataframe to use."""
        self.dh.load_profile_data(self.data_directory)
        self.dh.construct_dataframe(self.settings)

    def run_tool(self, tool='qc_smhi', return_layout=False):
        """Run the main QC tool (Bokeh)."""
        # TODO settings in yaml-files.. dynamic obj-load..
        plot = QCWorkTool(
            self.dh.df[self.settings.selected_keys],
            datasets=self.dh.raw_data,
            parameters=self.settings.data_parameters_with_units,
            plot_parameters_mapping=self.settings.plot_parameters_mapping,
            plot_keys=self.settings.plot_keys,
            color_fields=self.settings.q_colors,
            qflag_fields=self.settings.q_parameters,
            auto_q_flag_parameters=self.settings.q0_plot_keys,
            ctdpy_session=self.dh.ctd_session,
            multi_sensors=self.settings.multi_sensors,  # IMPORTANT!!! SMHI HAS MULTIPLE TEMP, SALT, DOXY SENSORS  # noqa: E501
            combo_plots=self.settings.combo_plots,
            user_download_directory=self.settings.user_download_directory,
        )
        plot.plot_stations()
        plot.plot_data()

        if return_layout:
            return plot.return_layout()



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-09-11 15:33

@author: a002028
"""
from bokeh.models import Div


def header_line():
    """Return Div object."""
    html_text = """
    <html>
    <body>
    <hr>
    </body>
    </html>
    """
    return Div(text=html_text)


def standard_block_header(text=None, width=300, height=40):
    """Return Div object."""
    html_text = """
    <style>
        body {
            text-align: left;
            vertical-align: text-top;
        }
        .centered {
            text-align: text-top;
            vertical-align: text-top;
        }
    </style>
    <div class="centered">
        <h4>""" + text + """</h4>
    </div>
    """
    return Div(text=html_text, width=width, height=height)


def get_info_block():
    """Return Div object."""
    text = """
    <h4>Info links</h4>
    <ul>
      <li><a href="https://docs.bokeh.org/en/latest/docs/user_guide/tools.html" target="_blank">Bokeh toolbar info</a></li>
      <li><a href="https://github.com/sharksmhi/sharkpylib/tree/master/sharkpylib/qc" target="_blank">SHARK-QC-library</a></li>
    </ul>

    <h4>QC routines</h4>
    <ol>
      <li>Range check</li>
      <li>Increase check</li>
      <li>Decrease check</li>
      <li>Sensor diff check</li>
      <li>Spike check</li>
    </ol>
    """  # noqa: E501
    return Div(text=text, width=200, height=100)


def get_export_info_block():
    """Return Div object."""
    text = """
    <h4>Download steps:</h4>
    <ol>
      <li>Select series using "map-lasso" or "Shift-table-select"</li>
      <li>Click on Download below</li>
    </ol>
    A folder with datafiles will be downloaded to your computer download-folder (eg. "Hämtade filer")
    """  # noqa: E501
    return Div(text=text)



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:24

@author: a002028
"""
from abc import ABC
import os

import pandas as pd
import numpy as np
from ctdpy.core.session import Session
from ctdpy.core.utils import generate_filepaths

from ctdvis.filter import Filter


def set_color_code(qf):
    """Return color based on the given flag."""
    if qf == 'B':
        return 'red'
    elif qf == 'S':
        return 'orange'
    elif qf == 'E':
        return 'green'
    else:
        return 'navy'


class Datadict(dict):
    """Data Dictionary.

    We intend to use filenames as keys with profile dataframes as items.
    """

    def __init__(self, data_directory=None):
        """Initiate."""
        super().__init__()
        self.data_directory = data_directory

    def load_data(self, return_session=False, filters=None):
        """Load CTD-standard-formnat data.

        Using ctdpy.

        Args:
            return_session (bool): True/False
            filters (dict | None): Filter out files to load based on month, serno or ship.
        """
        files = generate_filepaths(self.data_directory,
                                   pattern='ctd_profile_',
                                   endswith='.txt',
                                   only_from_dir=True)

        files = list(files)

        if filters:
            filter_obj = Filter([os.path.basename(f) for f in files])
            filter_obj.add_filter(**filters)
            files = [f for f in files if os.path.basename(f) in filter_obj.valid_file_names]

        ctd_session = Session(filepaths=files, reader='ctd_stdfmt')

        datasets = ctd_session.read()

        for key, item in datasets[0].items():
            self.append_item(key, item)

        if return_session:
            return ctd_session

    def append_item(self, key, item):
        """Append item to self."""
        print('Appending data for key: {}'.format(key))
        self.setdefault(key, item)


class Frame(pd.DataFrame, ABC):
    """Subclassing pandas.DataFrame.

    We intend to use Frame as a DataFrame for all selected profile dataset.
    """

    @property
    def _constructor(self):
        """Construct DataFrame, overides method in pd.DataFrame."""
        return Frame

    def convert_formats(self):
        """Convert hard coded formats."""
        self['SDATE'] = self[['YEAR', 'MONTH', 'DAY']].astype(str).apply('-'.join, axis=1)
        self['STIME'] = self[['HOUR', 'MINUTE']].astype(str).apply(':'.join, axis=1)

    def set_column_format(self, **kwargs):
        """Set dtype of each parameter in the dataframe."""
        for key, value in kwargs.items():
            if key in self.columns:
                try:
                    self[key] = self[key].astype(value)
                    if value == str:
                        self[key] = self[key].replace('nan', '')
                except ValueError:
                    self[key] = self[key].replace('', np.nan).astype(value)

    def add_color_columns(self, q_params, mapper=None):
        """Add color columns for each parameter."""
        for q_para in q_params:
            color_key = mapper.get(q_para)
            if q_para in self:
                self[color_key] = np.vectorize(set_color_code)(self[q_para].fillna(''))


class DataHandler:
    """Handler of data formats.

    Stores rawdata (ctd-standard-format).
    Appends selected filtered datasets (self.raw_data) to Frame (self.df).
    Conform self.df in order to use it in bokeh tools.
    """

    def __init__(self, filters):
        """Initiate."""
        self.filters = filters
        self.raw_data = Datadict()
        self.df = Frame(index=[])
        self.ctd_session = None

    def load_profile_data(self, directory):
        """Load data."""
        self.raw_data.data_directory = directory
        self.ctd_session = self.raw_data.load_data(return_session=True, filters=self.filters)

    def construct_dataframe(self, settings):
        """Set up dataframe according to a bokeh friendly format."""
        for key, item in self.raw_data.items():
            df = item['data'].copy()
            df['KEY'] = key.replace('ctd_profile_', '').replace('.txt', '')
            self.df = self.df.append(Frame(df))

        self.df.reset_index(drop=True, inplace=True)
        self.df.convert_formats()
        self.df.add_color_columns(settings.q_parameters, mapper=settings.q_colors_mapper)
        self.df.set_column_format(**settings.parameter_formats)
        self.check_columns(*settings.selected_keys)

    def check_columns(self, *args):
        """Add columns that are not present in the dataframe."""
        for col in args:
            if col not in self.df:
                self.df[col] = ''



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:33

@author: a002028
"""
from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass
from ctdvis import callbacks  # noqa: F401
from ctdvis import readers  # noqa: F401
from ctdvis import sources  # noqa: F401
from ctdvis import widgets  # noqa: F401

name = "ctdvis"



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-09-07 13:02

@author: a002028
"""
import numpy as np
from bokeh.models import ColumnDataSource


def setup_data_source(df, pmap=None, key_list=None, parameter_list=None):
    """Return a bokeh ColumnDataSource.

    Args:
        df: pandas.DataFrame
        pmap: Dictionary with parameter mapping
        key_list: array
        parameter_list: plot_keys and color_keys
    """
    main_source = {}
    for p in parameter_list:
        if p.startswith('COMNT'):
            main_source[p] = ['']
        elif p.startswith('color'):
            main_source[p] = ['navy']
        else:
            main_source[p] = [1]

    data_dict = {
        'main_source': ColumnDataSource(main_source),
        'default_source': ColumnDataSource(main_source)
    }

    for key in key_list:
        data_boolean = df['KEY'] == key
        key_df = df.loc[data_boolean, :]
        key_dict = {}
        for p in parameter_list:
            if p.startswith('COMNT'):
                df_column = p
            else:
                df_column = pmap.get(p) if not p.startswith('color') else p

            if df_column and df_column in key_df:
                key_dict[p] = key_df[df_column].values
            else:
                v = 'navy' if p.startswith('color') else np.nan
                key_dict[p] = [v] * key_df.__len__()

        data_dict[key] = ColumnDataSource(key_dict, name=key)

    return data_dict



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:25

@author: a002028
"""
from collections import Mapping
from datetime import datetime
from pyproj import CRS, transform
import numpy as np
import pandas as pd
import gsw
from matplotlib import colors
from matplotlib import cm


def get_color_palette(dep_serie=None):
    """Return palette."""
    number_of_colors = int(dep_serie.max() * 2 + 1)
    cm_map = cm.get_cmap('cool', number_of_colors)
    color_array = pd.Series([colors.to_hex(cm_map(c)) for c in range(number_of_colors)])
    return [color_array[int(d * 2)] if d > 0 else 0 for d in dep_serie]


def convert_projection(lats, lons):
    """Return position in google projection."""
    project_projection = CRS('EPSG:4326')
    google_projection = CRS('EPSG:3857')
    x, y = transform(project_projection, google_projection, lons, lats, always_xy=True)
    return x, y


def get_contour_arrays(x_min, x_max, y_min, y_max):
    """Calculate how many gridcells we need in the x and y dimensions.

    Assuming x_key = Salinity and y_key = Temperature
    """
    xdim = int(round((x_max - x_min) / 0.1 + 1, 0))
    ydim = int(round((y_max - y_min) / 0.1 + 1, 0))
    t_m = np.zeros((ydim, xdim))
    s_m = np.zeros((ydim, xdim))
    dens = np.zeros((ydim, xdim))
    ti = np.linspace(1, ydim - 1, ydim) * 0.1 + y_min
    si = np.linspace(1, xdim - 1, xdim) * 0.1 + x_min
    for j in range(0, int(ydim)):
        for i in range(0, int(xdim)):
            dens[j, i] = gsw.rho(si[i], ti[j], 0)
            s_m[j, i] = si[i]
            t_m[j, i] = ti[j]
    dens = dens
    return dens, t_m, s_m


def get_contour_data(x_min, x_max, y_min, y_max):
    """Return dictionary of contour data.

    Used for TS diagram.
    Salinity as sigma-T.

    Example:
        x_min, x_max, y_min, y_max = 0, 40, -10, 30
        data = get_contour_df(x_min, x_max, y_min, y_max)
    """
    dens, t_m, s_m = get_contour_arrays(x_min, x_max, y_min, y_max)
    dens = np.round(dens, 2)
    data = {}
    selected_densities = np.arange(-6, 31, 2) + 1000
    for s_dens in selected_densities:
        index = dens == s_dens
        data[str(s_dens)] = {'temp': t_m[index],
                             'salt': s_m[index],
                             'dens': dens[index]}
    return data


def get_time_as_format(now=None, timestamp=None, fmt=None):
    """Return time as string format."""
    if now:
        d = datetime.now()
    elif timestamp:
        raise NotImplementedError

    if fmt:
        return d.strftime(fmt)
    else:
        raise KeyError


def recursive_dict_update(d, u):
    """Recursive dictionary update.

    Copied from:
        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """  # noqa: E501
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d.setdefault(k, r)
        else:
            d.setdefault(k, u[k])
    return d



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:37

@author: a002028
"""
import os
import sys
import requests
from ctdvis.readers import JSONreader, load_json
try:
    from sharkpylib.qc.settings import Settings as shark_qc_settings
except ModuleNotFoundError as error:
    print(error)
    print('Could not import sharkpylib.qc.settings.',
          'Trying to catch parameter_dependencies via github..')
    shark_qc_settings = None


class Settings:
    """Setting class for CTDvis."""

    def __init__(self, visualize_setting=None):
        """Initiate."""
        self.multi_sensors = None
        self.combo_plots = None
        self.visualize_setting = visualize_setting or 'smhi_viz'
        self.base_directory = os.path.dirname(os.path.realpath(__file__))
        self.user_download_directory = os.path.join(os.path.expanduser("~"), "Downloads")
        etc_path = os.path.join(self.base_directory, 'etc')
        self._load_settings(etc_path)

        if shark_qc_settings:
            qc_settings = shark_qc_settings()
            self.parameter_dependencies = qc_settings.parameter_dependencies.copy()
        else:
            file_path = os.path.join(etc_path, 'parameter_dependencies.json')
            if not os.path.exists(file_path):
                try:
                    r = requests.get(
                        'https://raw.githubusercontent.com/sharksmhi/sharkpylib/master/sharkpylib/qc/etc/parameter_dependencies.json',  # noqa: E501
                        allow_redirects=True,
                    )
                    open(file_path, 'wb').write(r.content)
                    print('Download completed! file saved here: {}'.format(file_path))
                except ConnectionError as error:
                    raise error(
                        'Was not able to download https://raw.githubusercontent.com/sharksmhi/sharkpylib/master/sharkpylib/qc/etc/parameter_dependencies.json'  # noqa: E501
                        'and could not import sharkpylib.qc.settings\n '
                        'Try again when you have access to the internet!'
                    )
            self.parameter_dependencies = load_json(file_path)

    def __setattr__(self, name, value):
        """Define the setattr method for object self."""
        if name == 'dir_path':
            pass
        elif isinstance(value, str) and 'path' in name:
            name = os.path.join(self.base_directory, value)
        elif isinstance(value, dict) and 'paths' in name:
            self._check_for_paths(value)
        super().__setattr__(name, value)

    def _add_py_paths_to_system(self):
        """Append local python paths."""
        # TODO: Should not be needed.
        if hasattr(self, 'local_py_paths'):
            for item in self.local_py_paths.values():
                self._append_path_to_system(item)

    def _append_path_to_system(self, item):
        """Append local python paths."""
        # TODO: Should not be needed.
        if isinstance(item, str):
            if item not in sys.path:
                sys.path.append(item)
        elif isinstance(item, dict):
            for it in item.values():
                self._append_path_to_system(it)

    def _check_for_paths(self, dictionary):
        """Update the given dictionary with the ctdvis base folder.

        Since default path settings are set to ctdvis base folder
        we need to add that base folder to all paths.
        """
        for item, value in dictionary.items():
            if isinstance(value, dict):
                self._check_for_paths(value)
            elif 'path' in item:
                dictionary[item] = os.path.join(self.base_directory, value)

    def _load_settings(self, etc_path):
        """Load settings."""
        paths = self.generate_filepaths(etc_path, pattern='.json')
        path_list = []
        for p in paths:
            if 'vis_setup' in p:
                if self.visualize_setting in p:
                    path_list.append(p)
                else:
                    pass
            else:
                path_list.append(p)
        settings = JSONreader().load_json(config_files=path_list, return_dict=True)
        self.set_attributes(self, **settings)

    @staticmethod
    def set_attributes(obj, **kwargs):
        """Set attribute to the given object.

        With the possibility to add attributes to an object which is not 'self'.
        """
        # TODO Move to utils?
        for key, value in kwargs.items():
            setattr(obj, key, value)

    @staticmethod
    def generate_filepaths(directory, pattern=''):
        """Create generator for file paths."""
        # TODO Move to utils?
        for path, _, fids in os.walk(directory):
            for f in fids:
                if pattern in f:
                    yield os.path.abspath(os.path.join(path, f))

    @staticmethod
    def get_subdirectories(directory):
        """Return list of sub directories."""
        # TODO Move to utils?
        return [subdir for subdir in os.listdir(directory)
                if os.path.isdir(os.path.join(directory, subdir))]

    @staticmethod
    def get_filepaths_from_directory(directory):
        """Return list of file paths.

        Filter: only path of files, not folders.
        """
        # TODO Move to utils?
        return [os.path.join(directory, fid) for fid in os.listdir(directory)
                if not os.path.isdir(os.path.join(directory, fid))]

    @property
    def q_parameters(self):
        """Return Q-field-name of each parameter."""
        return [item.get('q_flag') for item in self.data_parameters.values()]

    @property
    def q0_parameters(self):
        """Return Q0-field-name of each parameter."""
        return [item.get('q0_flag') for item in self.data_parameters.values()]

    @property
    def q_colors(self):
        """Return color-field-name of each parameter."""
        return [item.get('plot_color_key') for item in self.data_parameters.values()]

    @property
    def q0_plot_keys(self):
        """Return Q0-plot-key of each parameter."""
        return [item.get('plot_q0_key') for item in self.data_parameters.values()]

    @property
    def plot_keys(self):
        """Return plot-key of each parameter."""
        return [item.get('plot_key') for item in self.data_parameters.values()]

    @property
    def data_parameters_with_units(self):
        """Return list of paramter names including there respective unit."""
        return [' '.join((key, item.get('unit'))) for key, item in self.data_parameters.items()]

    @property
    def selected_keys(self):
        """Return selected data keys."""
        return self.q_parameters + self.q0_parameters + self.q_colors + self.data_parameters_with_units + self.meta_parameters  # noqa: E501

    @property
    def q_colors_mapper(self):
        """Return mapper."""
        return {
            item.get('q_flag'): item.get('plot_color_key')
            for item in self.data_parameters.values()
        }

    @property
    def plot_parameters_mapping(self):
        """Return parameter mapping dictionary."""
        d = {}
        for key, item in self.data_parameters.items():
            d[item.get('plot_key')] = ' '.join((key, item.get('unit')))
            d[item.get('plot_q0_key')] = item.get('q0_flag')
            d[item.get('plot_color_key')] = item.get('q_flag')
            d[key] = {
                'q_flags': [
                    self.data_parameters[k].get('q_flag') for k in self.parameter_dependencies[key]
                    if k in self.data_parameters
                ],
                'color_keys': [
                    self.data_parameters[k].get('plot_color_key')
                    for k in self.parameter_dependencies[key] if k in self.data_parameters
                ]
            }
        return d

    @property
    def parameter_formats(self):
        """Return data type mapper."""
        d = {key: float for key in self.data_parameters_with_units}
        d.setdefault('COMNT_SAMP', str)
        return d


class InfoLog:
    """Basic information logger."""

    missing_stations = []

    def __init__(self, station=None):
        """Initiate."""
        if station:
            self.missing_stations.append(station)

    @classmethod
    def append_missing_station(cls, station):
        """Append station."""
        return cls('\t-' + station)

    @classmethod
    def print_missing_stations(cls):
        """Print missing station."""
        print('Missing stations:')
        print('\n'.join(cls.missing_stations))

    @classmethod
    def reset(cls):
        """Reset class.missing_stations."""
        cls.missing_stations = []


class ErrorCapturing:
    """Basic error logger."""

    errors = []

    def __init__(self, error=None):
        """Initiate."""
        if error:
            self.errors.append(error)

    @classmethod
    def append_error(cls, **error_kwargs):
        """Append error."""
        string = '\t-' + ', '.join(
            (': '.join((key, str(item))) for key, item in error_kwargs.items())
        )
        return cls(string)

    @classmethod
    def print_errors(cls):
        """Print error."""
        print('Errors:')
        print('\n'.join(cls.errors))

    @classmethod
    def reset(cls):
        """Reset class.errors."""
        cls.errors = []



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-03 11:49

@author: a002028
"""
from ctdvis.sources.ts import Source  # noqa: F401
from ctdvis.sources.data import setup_data_source  # noqa: F401



# Copyright (c) 2021 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2021-02-18 13:52

@author: johannes

"""
import os
import setuptools


requirements = []
with open('requirements.txt', 'r') as fh:
    for line in fh:
        requirements.append(line.strip())

NAME = 'ctdvis'
VERSION = "0.1.8"
README = open('READMEpypi.rst', 'r').read()

setuptools.setup(
    name=NAME,
    version=VERSION,
    author="SMHI - NODC",
    author_email="johannes.johansson@smhi.se",
    description="Quality control CTD/MVP data at the Swedish NODC",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/sharksmhi/ctdvis",
    packages=setuptools.find_packages(),
    package_data={'ctdvis': [
        os.path.join('etc', '*.json'),
        os.path.join('etc', 'vis_setup', '*.json'),
    ]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
)



# Copyright (c) 2020 SMHI, Swedish Meteorological and Hydrological Institute.
# License: MIT License (see LICENSE.txt or http://opensource.org/licenses/mit).
"""
Created on 2020-07-07 17:05

@author: a002028
"""
import pandas as pd


class Name:
    """Name properties object."""

    splitter = '_'

    def __init__(self, file_name):
        """Initiate."""
        args = file_name.replace('.txt', '').split(self.splitter)
        self.date = pd.Timestamp(args[2])
        self.shipc = args[3]
        self.serno = float(args[4])


class SplitNameList:
    """Name object handler."""

    dates = []
    ships = []
    sernos = []

    def __init__(self, name_list):
        """Initiate."""
        self.names = name_list
        for name in name_list:
            name_obj = Name(name)
            self.append_date(name_obj.date)
            self.append_shipc(name_obj.shipc)
            self.append_serno(name_obj.serno)

    @classmethod
    def append_date(cls, d):
        """Append date."""
        cls.dates.append(d)

    @classmethod
    def append_shipc(cls, s):
        """Append ship code."""
        cls.ships.append(s)

    @classmethod
    def append_serno(cls, s):
        """Append serie number."""
        cls.sernos.append(s)


class Filter:
    """Filter filenames according to ctd-standard format.

    Filename example: 'ctd_profile_20181208_34AR_0171.txt'.
    """

    def __init__(self, name_list):
        """Initiate."""
        lists = SplitNameList(name_list)
        self.serie_names = pd.Series(lists.names)
        self.serie_dates = pd.Series(lists.dates)
        self.serie_ships = pd.Series(lists.ships)
        self.serie_sernos = pd.Series(lists.sernos)

        self._boolean = True

    def add_filter(self, **kwargs):
        """Set boolean filter.

        If any valid filter arguments we append boolean according to:
        @boolean.setter (property.setter).
        """
        if 'month_list' in kwargs:
            print('month', self.serie_dates.dt.month)
            self.boolean = self.serie_dates.dt.month.isin(kwargs.get('month_list'))

        if 'ship_list' in kwargs:
            self.boolean = self.serie_ships.isin(kwargs.get('ship_list'))

        if 'serno_max' in kwargs:
            self.boolean = self.serie_sernos <= kwargs.get('serno_max')

        if 'serno_min' in kwargs:
            self.boolean = self.serie_sernos >= kwargs.get('serno_min')

    @property
    def valid_file_names(self):
        """Return valid filenames."""
        return self.serie_names[self.boolean].values

    @property
    def boolean(self):
        """Return boolean."""
        return self._boolean

    @boolean.setter
    def boolean(self, add_bool):
        """Set boolean."""
        self._boolean = self._boolean & add_bool
